# We construct a piecewise constant forward variance curve from variance swap data

# xi.curve <- function(expiries,w) function (t) 
# {
#   n <- length(w)
#   xi.vec <- c(w[1]/expiries[1], diff(w)/diff(expiries))
#   exp.vec <- c(0,expiries)
#   exp.vec[n+1] <- Inf # Long expiries are all in the last bucket
#   pos <- findInterval(t, vec = exp.vec)
#   res <- xi.vec[pos]
#   return(res)
# }

obj.w <- function(expiries,w.in)function(err.vec){
  
  w.in.1 <- w.in + 2*sqrt(w.in)*err.vec*sqrt(expiries)
  
  xi.vec <- c(w.in.1[1]/expiries[1],diff(w.in.1)/diff(expiries)) 
  
  dxi.dt <- diff(xi.vec)/diff(expiries)
  w.out <- c(0,cumsum(xi.vec[-1]*diff(expiries)))+xi.vec[1]*expiries[1]
  
  res <- sum((w.in-w.out)^2) + sum(dxi.dt^2)
  return(res*1e3)
}

xi.curve <- function(expiries,w.in,eps=0){
  
  n <- length(w.in)
  
  if(eps>0){
    res.optim <- optim(rep(0,n),obj.w(expiries,w.in),method="L-BFGS-B",
                     lower=rep(-eps,n),upper=rep(eps,n))
    err.vec <- res.optim$par
    w.in.1 <- w.in  + 2*sqrt(w.in)*err.vec*sqrt(expiries)}
  else {w.in.1 <- w.in}
  xi.vec.out <- c(w.in.1[1]/expiries[1],diff(w.in.1)/diff(expiries))
  xi.curve.raw <- function(t){
    res <- ifelse(t <= expiries[n], xi.vec.out[sum(expiries < t)+1],xi.vec.out[n])
    return(res)
  }
  xi.curve.out <- Vectorize(xi.curve.raw)
  fit.errs <- sqrt(w.in.1/expiries) - sqrt(w.in/expiries)
  
  return(list(xi.vec=xi.vec.out,xi.curve=xi.curve.out,fit.errs=fit.errs,w.out=w.in.1))
  
}