# Code to compute the rough Bergomi VIX^2 to VIX convexity 
# adjustment using the lognormal approximation

fH <- function(theta,h){
  integ <- function(x){((1+theta-x)^(h+1/2)-(1-x)^(h+1/2))^2}
  tmp <- integrate(integ, lower=0,upper=1)$value
  return(tmp/theta^2*(2*h)/(h+1/2)^2)
}

vixConvexity.raw <- function(H,nu,tau,delta=1/12){
  
  cH2 <- gamma(3/2-H)/(gamma(H+1/2)*gamma(2-2*H))
  etaNu <- 2*nu*sqrt(cH2)
  varLogPsi <- etaNu^2*tau^(2*H)*fH(delta/tau,H)
  convAdj <- exp(-varLogPsi/8)
  return(convAdj)
  
}

vixConvexity <- Vectorize(vixConvexity.raw,vectorize.args="tau")

eVix2.raw <- function(expiries,w){# w is total variance
  
  xi <- xi.curve(expiries,w,eps=0.000)$xi.curve
  function(tau){
    ev2 <- integrate(xi, lower = tau, upper = tau + 1/12)$value*12
    return(ev2)
  }
}

eVix2 <- function(expiries,w){Vectorize(eVix2.raw(expiries,w),vectorize.args="tau")}

# The following function takes a total variance curve as input and 
# outputs VIX futures prices for vixMaturities

f.vix <- function(expiries,w,vixMaturities)function (H, nu){
  conv.adj <- vixConvexity(H,nu,vixMaturities,delta=1/12)
  ev2 <- eVix2(expiries,w)(vixMaturities)
  sqrt(ev2) * conv.adj * 100
}
